/*
 * Copyright (c) 1989 Regents of the University of California.
 * All rights reserved.  The Berkeley software License Agreement
 * specifies the terms and conditions for redistribution.
 */

#ifndef lint
static char copyright[] = "Copyright (c) 1990 Regents of the University of California.\nAll rights reserved.\n";
static char SccsId[] = "@(#)@(#)pop_init.c      2.1  2.1 3/18/91";
#endif /* not lint */

#include <errno.h>
#include <stdio.h>
#include <sys/types.h>
#ifdef POPSCO
# include <sys/netinet/in.h>
#else
# include <netinet/in.h>
#endif
#include <netdb.h>
#include <arpa/inet.h>
#include <sys/param.h>

#if defined(SOLARIS2) || defined(SYSV) || defined(AIX)
#include        <string.h>
#define bcopy(src,dest,len)     (void) (memcpy(dest,src,len))
#define bzero(dest,len)         (void) (memset(dest, (char)NULL, len))
#define bcmp(b1,b2,n)           memcmp(b1,b2,n)
#ifndef index
# define index(s,c)             strchr(s,c)
#endif
#ifndef rindex
# define rindex(s,c)            strrchr(s,c)
#endif
#else
#include <strings.h>
#endif

#include "popper.h"

/* CNS Kerberos IV */
#ifdef KERBEROS
AUTH_DAT kdata;
#endif

extern int      errno;
/*
#ifdef POPSCO
extern struct state     _res;
#endif
*/

#ifdef STRDUP
#include <stddef.h>
#include <stdlib.h>

char *
strdup(str)
        char *str;
{
    int len;
    char *copy;

    len = strlen(str) + 1;
    if (!(copy = malloc((u_int)len)))
        return((char *)NULL);
    bcopy(str, copy, len);
    return(copy);
}
#endif

authenticate(p, addr)
     POP     *p;
     struct sockaddr_in *addr;
{

#ifdef KERBEROS
    Key_schedule schedule;
    KTEXT_ST ticket;
    char instance[INST_SZ];
    char version[9];
    int auth;

    if (p->kerberos) {
        strcpy(instance, "*");
        auth = krb_recvauth(0L, 0, &ticket, KERBEROS_SERVICE, instance,
                            addr, (struct sockaddr_in *) NULL,
                            &kdata, "", schedule, version);

        if (auth != KSUCCESS) {
            pop_msg(p, POP_FAILURE, "Kerberos authentication failure: %s",
                    krb_err_txt[auth]);
            pop_log(p, LOG_WARNING, "%s: (%s.%s@%s) %s", p->client,
                    kdata.pname, kdata.pinst, kdata.prealm, krb_err_txt[auth]);

            return(POP_FAILURE);
        }

# ifdef DEBUG
        if (p->debug)
            pop_log(p, POP_DEBUG, "%s.%s@%s (%s): ok", kdata.pname,
                kdata.pinst, kdata.prealm, inet_ntoa(addr->sin_addr));
# endif /* DEBUG */

        strncpy(p->user, kdata.pname, sizeof(p->user));

    }
#endif /* KERBEROS */

    return(POP_SUCCESS);
}


/* woju: standalone start */
#include <time.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/wait.h>
#include <sys/time.h>
#define PIDFILE  "/var/run/popper.pid"
#define TCP_QLEN        5
static struct sockaddr_in from;
static int rfds;                /* read file descriptor set */



/* ----------------------------------------------------- */
/* daemon : initialize and other stuff                   */
/* ----------------------------------------------------- */

start_daemon()
{
  int n, on;
  char buf[80];
  struct sockaddr_in fsin;

  /*
   * More idiot speed-hacking --- the first time conversion makes the C
   * library open the files containing the locale definition and time zone.
   * If this hasn't happened in the parent process, it happens in the
   * children, once per connection --- and it does add up.
   */

  time_t dummy = time(NULL);
  struct tm *dummy_time = localtime(&dummy);
  struct tm *other_dummy_time = gmtime(&dummy);
  strftime(buf, 80, "%d/%b/%Y:%H:%M:%S", dummy_time);

  n = getdtablesize();

  if (fork())
    exit(0);

  while (n)
    (void) close(--n);

  n = open(PIDFILE, O_WRONLY | O_CREAT | O_TRUNC, 0644);
  if (n >= 0)
  {
    sprintf(buf, "%5d\n", getpid());
    write(n, buf, 6);
    close(n);
  }

  (void) open("/dev/null", O_RDONLY);
  (void) dup2(0, 1);
  (void) dup2(0, 2);

  if ((n = open("/dev/tty", O_RDWR)) > 0)
  {
    ioctl(n, TIOCNOTTY, 0);
    close(n);
  }
/*
  openlog("ftpd", LOG_PID | LOG_ODELAY, LOG_AUTH);
*/
  openlog("popper", LOG_PID | LOG_NDELAY, POP_FACILITY);
  syslog(LOG_INFO, "start");
}


int
bind_port(port)
  int port;
{
  int sock, on;

  sock = socket(AF_INET, SOCK_STREAM, 0);
  if (sock < 0)
  {
    syslog(LOG_NOTICE, "socket\n");
    exit(1);
  }

  on = 1;
  if (setsockopt(sock, SOL_SOCKET, SO_REUSEADDR, (char *) &on, sizeof(on)) < 0)
    syslog(LOG_ERR, "(SO_REUSEADDR): %m");
  on = 0;

#if 0                           /* 0825 */
  on = 4096;
  setsockopt(sock, SOL_SOCKET, SO_SNDBUF, (char *) &on, sizeof(on));
  setsockopt(sock, SOL_SOCKET, SO_RCVBUF, (char *) &on, sizeof(on));
#endif

  from.sin_port = htons(port);
  if (bind(sock, (struct sockaddr *)&from, sizeof from) < 0) {
    syslog(LOG_INFO, "bind_port can't bind to %d",port);
    exit(1);
  }

  if (listen(sock, TCP_QLEN) < 0) {
    syslog(LOG_INFO, "bind_port can't listen to %d",port);
    exit(1);
  }

  FD_SET(sock, (fd_set *) & rfds);
  return sock;
}

void
reapchild()
{
  int state, pid;

  /* signal(SIGCHLD, reapchild); */
  while ((pid = waitpid(-1, &state, WNOHANG | WUNTRACED)) > 0);
}

/* ----------------------------------------------- */
/* o remote user name HPw                */
/* ----------------------------------------------- */

/*
 * rfc931() speaks a common subset of the RFC 931, AUTH, TAP, IDENT and RFC
 * 1413 protocols. It queries an RFC 931 etc. compatible daemon on a remote
 * host to look up the owner of a connection. The information should not be
 * used for authentication purposes. This routine intercepts alarm signals.
 *
 * Diagnostics are reported through syslog(3).
 *
 * Author: Wietse Venema, Eindhoven University of Technology, The Netherlands.
 */

#include <setjmp.h>

#define STRN_CPY(d,s,l) { strncpy((d),(s),(l)); (d)[(l)-1] = 0; }
#define STRING_LENGTH    60
#define RFC931_TIMEOUT   10
#define RFC931_PORT     113     /* Semi-well-known port */
#define ANY_PORT        0       /* Any old port will do */


/* ------------------------- */
/* timeout - handle timeouts */
/* ------------------------- */


char *inet_ntoa();
static jmp_buf timebuf;

static void
gtimeout(sig)
  int sig;
{
  longjmp(timebuf, sig);
}



void
getremotename(from, rhost, rname)
  struct sockaddr_in *from;
  char *rhost;
  char *rname;
{
  struct sockaddr_in our_sin;
  struct sockaddr_in rmt_sin;
  unsigned rmt_port, rmt_pt;
  unsigned our_port, our_pt;
  FILE *fp;
  char buffer[512], user[40], *cp;
  int s;
  struct hostent *hp;

  /* get remote host name */

  hp = NULL;
  if (setjmp(timebuf) == 0)
  {
    signal(SIGALRM, gtimeout);
    alarm(3);
    hp = gethostbyaddr((char *) &from->sin_addr, sizeof(struct in_addr),
      from->sin_family);
    alarm(0);
  }
  strncpy(rhost, hp ? hp->h_name : (char *) inet_ntoa(from->sin_addr), 39);
  rhost[39] = 0;

  /*
   * Use one unbuffered stdio stream for writing to and for reading from the
   * RFC931 etc. server. This is done because of a bug in the SunOS 4.1.x
   * stdio library. The bug may live in other stdio implementations, too.
   * When we use a single, buffered, bidirectional stdio stream ("r+" or "w+"
   * mode) we read our own output. Such behaviour would make sense with
   * resources that support random-access operations, but not with sockets.
   */

  s = sizeof our_sin;
  if (getsockname(0, (struct sockaddr*)&our_sin, &s) < 0)
    return;

  if ((s = socket(AF_INET, SOCK_STREAM, 0)) < 0)
  {
    perror("bbsd: socket in rfc931");
    return;
  }

  if (!(fp = fdopen(s, "r+")))
  {
    close(s);
    perror("bbsd:fdopen");
    return;
  }

  /*
   * Set up a timer so we won't get stuck while waiting for the server.
   */

  if (setjmp(timebuf) == 0)
  {
    signal(SIGALRM, gtimeout);
    alarm(RFC931_TIMEOUT);

    /*
     * Bind the local and remote ends of the query socket to the same IP
     * addresses as the connection under investigation. We go through all
     * this trouble because the local or remote system might have more than
     * one network address. The RFC931 etc. client sends only port numbers;
     * the server takes the IP addresses from the query socket.
     */

    our_pt = ntohs(our_sin.sin_port);
    our_sin.sin_port = htons(ANY_PORT);

    rmt_sin = *from;
    rmt_pt = ntohs(rmt_sin.sin_port);
    rmt_sin.sin_port = htons(RFC931_PORT);

    setbuf(fp, (char *) 0);
    s = fileno(fp);

    if (bind(s, (struct sockaddr *) & our_sin, sizeof(our_sin)) >= 0 &&
      connect(s, (struct sockaddr *) & rmt_sin, sizeof(rmt_sin)) >= 0)
    {

      /*
       * Send query to server. Neglect the risk that a 13-byte write would
       * have to be fragmented by the local system and cause trouble with
       * buggy System V stdio libraries.
       */

      fprintf(fp, "%u,%u\r\n", rmt_pt, our_pt);
      fflush(fp);

      /*
       * Read response from server. Use fgets()/sscanf() so we can work
       * around System V stdio libraries that incorrectly assume EOF when a
       * read from a socket returns less than requested.
       */

      if (fgets(buffer, sizeof(buffer), fp) && !ferror(fp) && !feof(fp)
        && sscanf(buffer, "%u , %u : USERID :%*[^:]:%39s",
          &rmt_port, &our_port, user) == 3
        && rmt_pt == rmt_port && our_pt == our_port)
      {

        /*
         * Strip trailing carriage return. It is part of the protocol, not
         * part of the data.
         */

        if (cp = strchr(user, '\r'))
          *cp = 0;
        strcpy(rname, user);
      }
    }
    alarm(0);
  }
  fclose(fp);
}

bdaemon(int argc, char** argv)
{
   int msock, nfds, csock, ofds, slot;
   struct timeval tv;
   pid_t pid;

   start_daemon();
   signal(SIGCHLD, reapchild);
   memset(&from, 0, sizeof(from));
   from.sin_family = AF_INET;
   rfds = 0;
   if (argc > 1) {
      msock = -1;
      for (nfds = 1; nfds < argc; nfds++) {
         csock = atoi(argv[nfds]);
         if (csock > 0)
            msock = bind_port(csock);
         else
            break;
      }
      if (msock < 0) {
         syslog(LOG_INFO, "bdaemon started with invalid arguments (no port)");
         exit(1);
      }
   }
   else {
      static int ports[] = {110 /*, 3456 , 4001 , 4002, 4003, 4004, 4005 */ };

      for (nfds = 0; nfds < sizeof(ports) / sizeof(int); nfds++)
      {
        msock = bind_port(ports[nfds]);
      }
   }


   /* main loop */

   ofds = rfds;
   nfds = msock + 1;

   for (;;) {
forever:
      rfds = ofds;
      tv.tv_sec = 60 * 30;
      tv.tv_usec = 0;
      msock = select(nfds, (fd_set *) & rfds, NULL, NULL, &tv);

      if (msock <=  0)
         continue;

      msock = 0;
      csock = 1;
      for (;;) {
         if (csock & rfds)
           break;
         if (++msock >= nfds)
           goto forever;
         csock <<= 1;
      }

      slot = sizeof(from);
      do {
         csock = accept(msock, (struct sockaddr *)&from, (int*)&slot);
      } while (csock < 0 && errno == EINTR);

      if (csock < 0)
        continue;

      pid = fork();

      if (!pid) {
         int len;
         struct sockaddr_in our;
         char rhost[40];
         char rname[40] = "?";
         char rfc931[200];
         int port;

         dup2(csock, 0);
         close(msock);
         getremotename(&from, rhost, rname);   /* RFC931 */
         sprintf(rfc931, "%s@%s", rname, rhost);
         syslog(LOG_INFO, "connect from %s", rfc931);

         setenv("REMOTEHOST", rhost, 1);
         setenv("REMOTENAME", rname, 1);
         setenv("RFC931", rfc931, 1);

/*
woju
*/
         len = sizeof(our);
         if (getsockname(0,(struct sockaddr *)&our,&len) < 0) {
            syslog(LOG_INFO, "getsockname: %s", strerror(errno));
            exit(1);
         }
/*
         connect_port = ntohs(our.sin_port);
*/

         while (--nfds >= 0)
            close(nfds);
         dup2(csock, 0);
         dup2(csock, 1);
         return;
      }
      else
         close(csock);
   }
   syslog(LOG_INFO, "NOTREACHED(bdaemon) reached!?");
/* NOTREACHED */
}


/*
 *  init:   Start a Post Office Protocol session
 */

pop_init(p, argcount, argmessage)
POP     *       p;
int             argcount;
char    **      argmessage;
{

    struct sockaddr_in      cs;                 /*  Communication parameters */
    struct hostent      *   ch;                 /*  Client host information */
    int                     errflag = 0;
    int                     c;
    int                     len;
    extern char         *   optarg;
    int                     options = 0;
    int                     sp = 0;             /*  Socket pointer */
    char                *   trace_file_name;
    struct hostent      *   hp = NULL;

    /*  Initialize the POP parameter block */
    bzero ((char *)p,(int)sizeof(POP));

    /*  Initialize maildrop status variables in the POP parameter block */
    p->msgs_deleted = 0;
    p->last_msg = 0;
    p->bytes_deleted = 0;
    p->drop_size = 0;
    p->mmdf_separator = NULL;
    p->bulldir = BULLDIR;
    p->dirty = 0;
    p->kerberos = 0;
#ifdef SERVER_MODE
    p->server_mode = 1;
#else
    p->server_mode = 0;
#endif

    /*  Save my name in a global variable */
    p->myname = argmessage[0];

    /*  Get the name of our host */
    if ((p->myhost = (char *)malloc(MAXHOSTNAMELEN+1)) == NULL) {
        perror("malloc");
        exit(1);
    }
    (void)gethostname(p->myhost,MAXHOSTNAMELEN);
    if (hp = gethostbyname(p->myhost)) {
        if (! index(hp->h_name, '.')) {         /* FQN not returned */
            /*
             * SVR4 resolver is stupid and returns h_name as whatever
             * you gave gethostbyname.  Thus do a reverse lookup
             * on the first address and hope for the best.
             */
            u_long x = *(u_long *)hp->h_addr_list[0];
            if ((hp = gethostbyaddr((char *)&x, 4, AF_INET)) != NULL) {
                (void) strncpy (p->myhost, hp->h_name, MAXHOSTNAMELEN);
                p->myhost[MAXHOSTNAMELEN] = '\0';
            }
        }
        else {
            (void) strncpy (p->myhost, hp->h_name, MAXHOSTNAMELEN);
            p->myhost[MAXHOSTNAMELEN] = '\0';
        }
    }

#if 0
woju: standalone
    /*  Open the log file */
#ifdef SYSLOG42
    (void)openlog(p->myname,0);
#else
    (void)openlog(p->myname,POP_LOGOPTS,POP_FACILITY);
#endif

    /*  Process command line arguments */
    while ((c = getopt(argcount,argmessage,"dkst:T:b:")) != EOF)
        switch (c) {

            /*  Debugging requested */
            case 'd':
                p->debug++;
                options |= SO_DEBUG;
#ifdef DEBUG
                if (p->debug)
                    pop_log(p,POP_PRIORITY,"Debugging turned on");
#endif
                break;

#ifdef KERBEROS
            case 'k':
                p->kerberos++;
            break;
#endif  /* KERBEROS */

            /* Stats requested */
            case 's':
                p->stats++;
                break;

            /* Bulletins requested */
            case 'b':
                p->bulldir = optarg;
                break;

            /*  Debugging trace file specified */
            case 't':
                p->debug++;

#ifdef DEBUG
                if (p->trace)
                    pop_log(p,POP_PRIORITY,
                    "Tracing session and debugging information in file \"%s\"",
                            trace_file_name);
#endif

                if ((p->trace = fopen(optarg,"a+")) == NULL) {
                    pop_log(p,POP_PRIORITY,
                        "Unable to open trace file \"%s\", err = %d",
                            optarg,errno);
                    exit(1);
                }
                trace_file_name = optarg;
                break;

            /*  Timeout value passed.  Default changed */
            case 'T':
                pop_timeout = atoi(optarg);
                break;

            /*  Unknown option received */
            default:
                errflag++;
        }

    /*  Exit if bad options specified */
    if (errflag) {
#ifdef KERBEROS
        (void)fprintf(stderr,"Usage: %s [-d] [-k] [-s] [-t trace-file] [-T timeout] [-b bulldir]\n",argmessage[0]);
#else
        (void)fprintf(stderr,"Usage: %s [-d] [-s] [-t trace-file] [-T timeout] [-b bulldir]\n",argmessage[0]);
#endif
        exit(1);
    }
woju: standalone
#endif
    p->stats++;
    bdaemon(argcount, argmessage);

    /*  Get the address and socket of the client to whom I am speaking */
    len = sizeof(cs);
    if (getpeername(sp,(struct sockaddr *)&cs,&len) < 0){
        pop_log(p,POP_PRIORITY,
            "Unable to obtain socket and address of client, err = %d",errno);
        exit(1);
    }

    /*  Save the dotted decimal form of the client's IP address
        in the POP parameter block */
    p->ipaddr = (char *)strdup(inet_ntoa(cs.sin_addr));

    /*  Save the client's port */
    p->ipport = ntohs(cs.sin_port);

    /*  Get the canonical name of the host to whom I am speaking */
    ch = gethostbyaddr((char *) &cs.sin_addr, sizeof(cs.sin_addr), AF_INET);
    if (ch == NULL){
        pop_log(p,POP_PRIORITY,
            "(v%s) Unable to get canonical name of client, err = %d",
            VERSION, errno);
        p->client = p->ipaddr;
    }
    /*  Save the cannonical name of the client host in
        the POP parameter block */
    else {

#ifndef BIND43
        p->client = (char *)strdup(ch->h_name);
#else

#if 0
woju
# ifndef SCOR5
#       include <arpa/nameser.h>
#       include <resolv.h>
# endif
#endif
        /*  Distrust distant nameservers */

#if !(defined(BSD) && (BSD >= 199103)) && !defined(OSF1) && !defined(HPUX10)
# if (!defined(__RES)) || (__RES < 19940415)
#  ifdef SCOR5
        extern struct __res_state       _res;
#  else
        extern struct state     _res;
#  endif
# endif
#endif
        struct hostent      *   ch_again;
        char            *   *   addrp;
        char                    h_name[MAXHOSTNAMELEN + 1];

        /*  We already have a fully-qualified name */
#ifdef RES_DEFNAMES
        _res.options &= ~RES_DEFNAMES;
#endif

        strncpy(h_name, ch->h_name, sizeof(h_name));

        /*  See if the name obtained for the client's IP
            address returns an address */
        if ((ch_again = gethostbyname(h_name)) == NULL) {
            pop_log(p,POP_PRIORITY,
                "Client at \"%s\" resolves to an unknown host name \"%s\"",
                    p->ipaddr, h_name);
            p->client = p->ipaddr;
        }
        else {
            /*  Save the host name (the previous value was
                destroyed by gethostbyname) */
            p->client = (char *)strdup(ch_again->h_name);

            /*  Look for the client's IP address in the list returned
                for its name */
            for (addrp=ch_again->h_addr_list; *addrp; ++addrp)
                if (bcmp(*addrp,&(cs.sin_addr),sizeof(cs.sin_addr)) == 0) break;

            if (!*addrp) {
                pop_log (p,POP_PRIORITY,
                    "Client address \"%s\" not listed for its host name \"%s\"",
                        p->ipaddr,h_name);
                p->client = p->ipaddr;
            }
        }

#ifdef RES_DEFNAMES
        /*
         *  Must restore nameserver options since code in crypt uses
         *  gethostbyname call without fully qualified domain name!
         */
        _res.options |= RES_DEFNAMES;
#endif

#endif /* BIND43 */
    }

    /*  Create input file stream for TCP/IP communication */
    if ((p->input = fdopen(sp,"r")) == NULL){
        pop_log(p,POP_PRIORITY,
            "Unable to open communication stream for input, err = %d",errno);
        exit (1);
    }

    /*  Create output file stream for TCP/IP communication */
    if ((p->output = fdopen(sp,"w")) == NULL){
        pop_log(p,POP_PRIORITY,
            "Unable to open communication stream for output, err = %d",errno);
        exit (1);
    }

#ifdef DEBUG
    if (p->debug)
        pop_log(p,POP_PRIORITY,
            "(v%s) Servicing request from \"%s\" at %s",
                VERSION,p->client,p->ipaddr);
#endif

    return(authenticate(p, &cs));
}

